'''
See COPYRIGHT.md for copyright information.
'''
import xml.dom, datetime
import regex as re
from arelle import XbrlConst, XbrlUtil, XmlUtil
from arelle.formula import XPathContext
from arelle.ModelObject import ModelObject, ModelAttribute
from arelle.ModelValue import qname, QName, dateTime, DATE, DATETIME, DATEUNION, DateTime, dateUnionEqual, anyURI
from arelle.FunctionUtil import anytypeArg, stringArg, numericArg, qnameArg, nodeArg, atomicArg
from arelle.ModelXbrl import ModelXbrl
from arelle.ModelDtsObject import anonymousTypeSuffix, ModelConcept
from arelle.ModelInstanceObject import ModelDimensionValue, ModelFact, ModelInlineFact
from arelle.ModelFormulaObject import ModelFormulaResource
from arelle.PythonUtil import flattenSequence
from arelle.formula.XPathParser import OperationDef
from arelle.XmlValidateConst import UNKNOWN, VALID, VALID_NO_CONTENT
from arelle.XmlValidate import validate as xmlValidate, NCNamePattern
from arelle.ValidateXbrlCalcs import inferredDecimals, inferredPrecision
from arelle.ValidateXbrlDimensions import priItemElrHcRels
from arelle.Locale import format_picture
from lxml import etree
from math import isnan, isinf

class xfiFunctionNotAvailable(Exception):
    def __init__(self):
        self.args =  (_("xfi function not available"),)
    def __repr__(self):
        return self.args[0]

def call(
        xc: XPathContext.XPathContext,
        p: OperationDef,
        localname: str,
        args: XPathContext.ResultStack,
) -> XPathContext.RecursiveContextItem:
    try:
        if localname not in xfiFunctions: raise xfiFunctionNotAvailable
        return xfiFunctions[localname](xc, p, args)
    except xfiFunctionNotAvailable:
        raise XPathContext.FunctionNotAvailable("xfi:{0}".format(localname))

def instance(xc, p, args, i=0):
    if i >= len(args):  # missing argument means to use the standard input instance
        return xc.modelXbrl
    if len(args[i]) != 1: # a sequence of instances isn't acceptable to these classes of functions
        raise XPathContext.FunctionArgType(i+1,"xbrl:xbrl")
    xbrliXbrl = anytypeArg(xc, args, i, "xbrli:xbrl")
    if isinstance(xbrliXbrl, ModelObject) and xbrliXbrl.elementQname == XbrlConst.qnXbrliXbrl:
        return xbrliXbrl.modelXbrl
    elif isinstance(xbrliXbrl, ModelXbrl):
        return xbrliXbrl
    raise XPathContext.FunctionArgType(i,"xbrl:xbrl")

def item(xc, args, i=0):
    if len(args[i]) != 1: raise XPathContext.FunctionArgType(i+1,"xbrl:item")
    modelItem = xc.modelItem(args[i][0])
    if modelItem is not None:
        return modelItem
    raise XPathContext.FunctionArgType(i,"xbrl:item")

def xbrlTuple(xc, args, i=0):
    # can't name this just tuple because then it hides tuple() constructor of Python
    if len(args[i]) != 1: raise XPathContext.FunctionArgType(i+1,"xbrl:tuple")
    modelTuple = args[i][0]
    if isinstance(modelTuple, (ModelFact, ModelInlineFact)) and modelTuple.isTuple:
        return modelTuple
    raise XPathContext.FunctionArgType(i,"xbrl:tuple")

def item_context(xc, args, i=0):
    return item(xc, args, i).context

def item_context_element(xc, args, name):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        return XmlUtil.descendant(context, XbrlConst.xbrli, name)
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def context(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    return item_context(xc, args)

def unit(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:item")
    modelItem = xc.modelItem(args[0][0])
    if modelItem is not None:
        modelConcept = modelItem.concept
        if modelConcept.isNumeric and not modelConcept.isFraction:
            return modelItem.unit
        return []
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def unit_numerator(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:unit")
    unit = args[0][0]
    if isinstance(unit,ModelObject) and \
       unit.localName == "unit" and unit.namespaceURI == XbrlConst.xbrli:
        measuresParent = XmlUtil.descendant(unit, XbrlConst.xbrli, "unitNumerator")
        if measuresParent is None: measuresParent = unit
        return XmlUtil.descendants(measuresParent, XbrlConst.xbrli, "measure")
    raise XPathContext.FunctionArgType(1,"xbrl:unit")

def unit_denominator(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:unit")
    unit = args[0][0]
    if isinstance(unit,ModelObject) and \
       unit.localName == "unit" and unit.namespaceURI == XbrlConst.xbrli:
        measuresParent = XmlUtil.descendant(unit, XbrlConst.xbrli, "unitDenominator")
        if measuresParent is None: return []
        return XmlUtil.descendants(measuresParent, XbrlConst.xbrli, "measure")
    raise XPathContext.FunctionArgType(1,"xbrl:unit")

def measure_name(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:measure")
    unit = args[0][0]
    if isinstance(unit,ModelObject) and \
       unit.localName == "measure" and unit.namespaceURI == XbrlConst.xbrli:
        return qname(unit, XmlUtil.text(unit))
    raise XPathContext.FunctionArgType(1,"xbrl:unit")

def period(xc, p, args):
    return item_context_element(xc, args, "period")

def context_period(xc, p, args):
    return parent_child(args, "context", "period")

def parent_child(args, parentName, childName, findDescendant=False):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:" + parentName)
    parent = args[0][0]
    if isinstance(parent,ModelObject) and \
       parent.localName == parentName and parent.namespaceURI == XbrlConst.xbrli:
        if childName.startswith('@'):
            return parent.get(childName[1:])
        elif childName == 'text()':
            return XmlUtil.textNotStripped(parent)
        elif childName == 'strip-text()':
            return XmlUtil.text(parent)
        elif findDescendant:
            return XmlUtil.descendant(parent, XbrlConst.xbrli, childName)
        else:
            return XmlUtil.child(parent, XbrlConst.xbrli, childName)
    raise XPathContext.FunctionArgType(1,"xbrl:" + parentName)

def is_start_end_period(xc, p, args):
    return is_period_type(args, "startDate")

def is_forever_period(xc, p, args):
    return is_period_type(args, "forever")

def is_duration_period(xc, p, args):
    return is_period_type(args, ("forever","startDate") )

def is_instant_period(xc, p, args):
    return is_period_type(args, "instant")

def is_period_type(args, periodElement):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:period")
    period = args[0][0]
    if isinstance(period,ModelObject) and \
       period.localName == "period" and period.namespaceURI == XbrlConst.xbrli:
        return XmlUtil.hasChild(period, XbrlConst.xbrli, periodElement)
    raise XPathContext.FunctionArgType(1,"xbrl:period")

def period_start(xc, p, args):
    return period_datetime(p, args, ("startDate","instant"))

def period_end(xc, p, args):
    return period_datetime(p, args, ("endDate","instant"))

def period_instant(xc, p, args):
    return period_datetime(p, args, "instant")

def period_datetime(p, args, periodElement):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:period")
    period = args[0][0]
    if (isinstance(period,ModelObject) == 1 and
        period.localName == "period" and period.namespaceURI == XbrlConst.xbrli):
        child = XmlUtil.child(period, XbrlConst.xbrli, periodElement)
        if child is not None:
            addOneDay = child.localName != "startDate"
            return dateTime( child, addOneDay=addOneDay, type=DATETIME)
        elif periodElement == "instant":
            raise XPathContext.XPathException(p, 'xfie:PeriodIsNotInstant', _('Period is not instant'))
        else:
            raise XPathContext.XPathException(p, 'xfie:PeriodIsForever', _('Period is forever'))
    raise XPathContext.FunctionArgType(1,"xbrl:period")

def entity(xc, p, args):
    return item_context_element(xc, args, "entity")

def context_entity(xc, p, args):
    return parent_child(args, "context", "entity")

def identifier(xc, p, args):
    return item_context_element(xc, args, "identifier")

def context_identifier(xc, p, args):
    return parent_child(args, "context", "identifier", True)

def entity_identifier(xc, p, args):
    return parent_child(args, "entity", "identifier")

def identifier_value(xc, p, args):
    return parent_child(args, "identifier", "strip-text()")

def identifier_scheme(xc, p, args):
    scheme = parent_child(args, "identifier", "@scheme")
    if scheme is None:
        return None
    return anyURI(scheme)

def fact_identifier_value(xc, p, args):
    return XmlUtil.text(item_context_element(xc, args, "identifier")).strip()

def fact_identifier_scheme(xc, p, args):
    scheme = item_context_element(xc, args, "identifier").get("scheme")
    if scheme is None:
        return None
    return anyURI(scheme)

def segment(xc, p, args):
    seg = item_context_element(xc, args, "segment")
    if seg is None:
        return () # no segment
    return seg

def entity_segment(xc, p, args):
    seg = parent_child(args, "entity", "segment")
    if seg is None:
        return () # no segment
    return seg

def context_segment(xc, p, args):
    seg = parent_child(args, "context", "segment", True)
    if seg is None:
        return () # no segment
    return seg

def scenario(xc, p, args):
    scen = item_context_element(xc, args, "scenario")
    if scen is None:
        return () # no segment
    return scen

def context_scenario(xc, p, args):
    scen = parent_child(args, "context", "scenario")
    if scen is None:
        return () # no segment
    return scen

def precision(xc, p, args):
    return infer_precision_decimals(xc, p, args, "precision")

def decimals(xc, p, args):
    return infer_precision_decimals(xc, p, args, "decimals")

def infer_precision_decimals(xc, p, args, attrName):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    if len(args[0]) != 1: raise XPathContext.FunctionArgType(1,"xbrl:item",args[0])
    modelItem = xc.modelItem(args[0][0])
    if modelItem is None:
        raise XPathContext.FunctionArgType(1,"xbrl:item")
    modelConcept = modelItem.concept
    if modelConcept.isFraction:
        return 'INF'
    if modelConcept.isNumeric:
        p = inferredPrecision(modelItem) if attrName == "precision" else inferredDecimals(modelItem)
        if isinf(p):
            return 'INF'
        if isnan(p):
            raise XPathContext.XPathException(p, 'xfie:ItemIsNotNumeric', _('Argument 1 {0} is not inferrable.').format(attrName))
        return p
    raise XPathContext.XPathException(p, 'xfie:ItemIsNotNumeric', _('Argument 1 is not reported with {0}.').format(attrName))

def numeric(xc, p, args):
    return conceptProperty(xc, p, args, "numeric")

def non_numeric(xc, p, args):
    return conceptProperty(xc, p, args, "non-numeric")

def fraction(xc, p, args):
    return conceptProperty(xc, p, args, "fraction")

def conceptProperty(xc, p, args, property):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    qn = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    if qn:
        modelConcept = xc.modelXbrl.qnameConcepts.get(qn)
        if modelConcept is not None:
            if property == "numeric": return modelConcept.isNumeric or modelConcept.isFraction
            if property == "non-numeric": return modelConcept.isItem and not (modelConcept.isNumeric or modelConcept.isFraction)
            if property == "fraction": return modelConcept.isFraction
    return False

def checkXffFunctionUse(xc, p, functionName):
    # check function use after checking argument types
    if xc.progHeader is not None and xc.progHeader.element is not None:
        try:
            modelResourceElt = xc.progHeader.element._modelResourceElt
        except AttributeError:
            modelResourceElt = xc.progHeader.element
            while (modelResourceElt is not None and not isinstance(modelResourceElt, ModelFormulaResource)):
                modelResourceElt = modelResourceElt.getparent()
            xc.progHeader.element._modelResourceElt = modelResourceElt

        if (modelResourceElt is None or
            modelResourceElt.localName not in ("formula", "consistencyAssertion", "valueAssertion", "precondition", "message")):
            raise XPathContext.XPathException(p, 'xffe:invalidFunctionUse', _('Function xff:uncovered-aspect cannot be used on an XPath expression associated with a {0}').format(xc.progHeader.element.localName))

    if xc.variableSet is not None and xc.variableSet.implicitFiltering  == "false":
        raise XPathContext.XPathException(p, 'xffe:invalidFunctionUse', _('Function xff:uncovered-aspect cannot be used with implicitFiltering=false'))

def uncovered_aspect(xc, p, args):
    from arelle.ModelFormulaObject import aspectFromToken, Aspect
    from arelle.formula.FormulaEvaluator import uncoveredAspectValue
    if len(args) not in (1,2): raise XPathContext.FunctionNumArgs()
    aspect = aspectFromToken.get(stringArg(xc, args, 0, "xs:token").strip())
    if aspect == Aspect.DIMENSIONS:
        qn = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)

    checkXffFunctionUse(xc, p, "uncovered-aspect")

    if aspect == Aspect.DIMENSIONS:
        if qn:
            modelConcept = xc.modelXbrl.qnameConcepts.get(qn)
            if modelConcept is not None and modelConcept.isDimensionItem:
                aspect = qn
            else:
                return ()   # not a dimension
            dimValue = uncoveredAspectValue(xc, aspect)
            if isinstance(dimValue, ModelDimensionValue):
                if dimValue.isExplicit:
                    return dimValue.memberQname
                elif dimValue.isTyped:
                    return dimValue     # return the typedMember element, not its contents
            elif isinstance(dimValue, QName): # qname for explicit or node for typed
                return dimValue
            return ()
    aspectValue = uncoveredAspectValue(xc, aspect)
    if aspectValue is None:
        return ()
    return aspectValue

def has_fallback_value(xc, p, args):
    from arelle.formula.FormulaEvaluator import variableBindingIsFallback
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    variableQname = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)

    checkXffFunctionUse(xc, p, "has-fallback-value")

    return variableBindingIsFallback(xc, variableQname)

def uncovered_non_dimensional_aspects(xc, p, args):
    return uncovered_aspects(xc, p, args, dimensionAspects=False)

def uncovered_dimensional_aspects(xc, p, args):
    return uncovered_aspects(xc, p, args, dimensionAspects=True)

def uncovered_aspects(xc, p, args, dimensionAspects=False):
    from arelle.ModelFormulaObject import aspectToToken, Aspect
    from arelle.formula.FormulaEvaluator import uncoveredVariableSetAspects
    if len(args) != 0: raise XPathContext.FunctionNumArgs()

    # check function use after checking argument types
    if xc.progHeader is not None and xc.progHeader.element is not None:
        if xc.progHeader.element.localName not in ("formula", "consistencyAssertion", "valueAssertion", "message"):
            raise XPathContext.XPathException(p, 'xffe:invalidFunctionUse', _('Function xff:uncovered-aspect cannot be used on an XPath expression associated with a {0}').format(xc.progHeader.element.localName))
        if xc.variableSet is not None and xc.variableSet.implicitFiltering  == "false":
            raise XPathContext.XPathException(p, 'xffe:invalidFunctionUse', _('Function xff:uncovered-aspect cannot be used with implicitFiltering=false'))

    uncoveredAspects = uncoveredVariableSetAspects(xc)
    return [(a if dimensionAspects else aspectToToken.get(a))
            for a in uncoveredAspects if a != Aspect.DIMENSIONS and isinstance(a,QName) == dimensionAspects ]

def nodesEqual(xc, args, test, mustBeItems=False, nonItemErrCode=None):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    seq1 = flattenSequence(args[0])
    seq2 = flattenSequence(args[1])
    for i, node1 in enumerate(seq1):
        try:
            node2 = seq2[i]
            if not isinstance(node1, (ModelObject,ModelAttribute)):
                raise XPathContext.FunctionArgType(1,"node()*")
            if not isinstance(node2, (ModelObject,ModelAttribute)):
                raise XPathContext.FunctionArgType(2,"node()*")
            if mustBeItems:
                if not isinstance(node1, (ModelFact, ModelInlineFact)) or not node1.isItem:
                    raise XPathContext.FunctionArgType(1,"xbrl:item*", errCode=nonItemErrCode)
                if not isinstance(node2, (ModelFact, ModelInlineFact)) or not node2.isItem:
                    raise XPathContext.FunctionArgType(2,"xbrl:item*", errCode=nonItemErrCode)
            if not test(node1, node2):
                return False
        except IndexError:
            return False
    return True

def setsEqual(xc, args, test, mustBeItems=False):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    seq1 = flattenSequence(args[0])
    seq2 = flattenSequence(args[1])
    for node1 in seq1:
        if not isinstance(node1, ModelObject):
            raise XPathContext.FunctionArgType(1,"node()*")
        if mustBeItems and (not isinstance(node1, (ModelFact, ModelInlineFact)) or not node1.isItem):
            raise XPathContext.FunctionArgType(1,"xbrl:item*", errCode="xfie:NodeIsNotXbrlItem")
    for node2 in seq2:
        if not isinstance(node2, ModelObject):
            raise XPathContext.FunctionArgType(2,"node()*")
        if mustBeItems and (not isinstance(node2, (ModelFact, ModelInlineFact)) or not node2.isItem):
            raise XPathContext.FunctionArgType(2,"xbrl:item*", errCode="xfie:NodeIsNotXbrlItem")
    if len(set(seq1)) != len(set(seq2)): # sequences can have nondistinct duplicates, just same set lengths needed
        return False
    for node1 in seq1:
        if not any(test(node1, node2) for node2 in seq2):
            return False
    return True

def identical_nodes(xc, p, args):
    return nodesEqual(xc, args, identical_nodes_test)

def identical_nodes_test(node1, node2):
    return node1 == node2

def s_equal(xc, p, args):
    return nodesEqual(xc, args, s_equal_test)

def s_equal_test(node1, node2):
    if (isinstance(node1, (ModelFact, ModelInlineFact)) and node1.isItem and
        isinstance(node2, (ModelFact, ModelInlineFact)) and node2.isItem):
        return (c_equal_test(node1, node2) and u_equal_test(node1, node2) and
                XbrlUtil.xEqual(node1, node2) and
                # must be validated (by xEqual) before precision tests to assure xAttributes is set
                node1.xAttributes.get("precision") == node2.xAttributes.get("precision") and
                node1.xAttributes.get("decimals") == node2.xAttributes.get("decimals"))
    elif isinstance(node1, ModelObject):
        if isinstance(node2, ModelObject):
            return XbrlUtil.sEqual(node1.modelXbrl, node1, node2, excludeIDs=XbrlUtil.TOP_IDs_EXCLUDED, dts2=node2.modelXbrl)
        else:
            return False
    elif isinstance(node1, ModelAttribute):
        if isinstance(node2, ModelAttribute):
            return node1.text == node2.text
    return False

def u_equal(xc, p, args):
    return nodesEqual(xc, args, u_equal_test, mustBeItems=True, nonItemErrCode="xfie:NodeIsNotXbrlItem")

def u_equal_test(modelItem1, modelItem2):
    modelUnit1 = modelItem1.unit
    modelUnit2 = modelItem2.unit
    if modelUnit1 is None:
        return modelUnit2 is None
    else:
        return modelUnit1.isEqualTo(modelUnit2)

def v_equal(xc, p, args):
    return nodesEqual(xc, args, v_equal_test, mustBeItems=True, nonItemErrCode="xfie:NodeIsNotXbrlItem")

def v_equal_test(modelItem1, modelItem2):
    return modelItem1.isVEqualTo(modelItem2)

def c_equal(xc, p, args):
    return nodesEqual(xc, args, c_equal_test, mustBeItems=True, nonItemErrCode="xfie:NodeIsNotXbrlItem")

def c_equal_test(modelItem1, modelItem2):
    modelCntx1 = modelItem1.context
    modelCntx2 = modelItem2.context
    if modelCntx1 is None:
        return modelCntx2 is None
    else:
        return modelCntx1.isEqualTo(modelCntx2,dimensionalAspectModel=False)

def identical_node_set(xc, p, args):
    return setsEqual(xc, args, identical_nodes_test)

def s_equal_set(xc, p, args):
    return setsEqual(xc, args, s_equal_test)

def v_equal_set(xc, p, args):
    return setsEqual(xc, args, v_equal_test, mustBeItems=True)

def c_equal_set(xc, p, args):
    return setsEqual(xc, args, c_equal_test, mustBeItems=True)

def u_equal_set(xc, p, args):
    return setsEqual(xc, args, u_equal_test, mustBeItems=True)

def x_equal(xc, p, args):
    return nodesEqual(xc, args, x_equal_test)

def x_equal_test(node1, node2):
    if isinstance(node1, ModelObject):
        if isinstance(node2, ModelObject):
            return XbrlUtil.xEqual(node1, node2)
        else:
            return False
    elif isinstance(node1, ModelAttribute):
        if isinstance(node2, ModelAttribute):
            return node1.sValue == node2.sValue
    return False


def duplicate_item(xc, p, args):
    node1 = item(xc, args, 0)
    node2 = item(xc, args, 1)
    if node1.isItem and node2.isItem:
        return node1.isDuplicateOf(node2)
    return False

def duplicate_tuple(xc, p, args):
    node1 = xbrlTuple(xc, args, 0)
    node2 = xbrlTuple(xc, args, 1)
    return duplicate_tuple_test(node1, node2)

def duplicate_tuple_test(node1, node2, topLevel=True):
    if node1.isTuple and node2.isTuple:
        return node1.isDuplicateOf(node2)
    return False

def p_equal(xc, p, args):
    return nodesEqual(xc, args, p_equal_test)

def p_equal_test(node1, node2):
    if not isinstance(node1, (ModelFact, ModelInlineFact)) or not (node1.isItem or node1.isTuple):
        raise XPathContext.FunctionArgType(1,"xbrli:item or xbrli:tuple", errCode="xfie:ElementIsNotXbrlConcept")
    if not isinstance(node2, (ModelFact, ModelInlineFact)) or not (node1.isItem or node1.isTuple):
        raise XPathContext.FunctionArgType(2,"xbrli:item or xbrli:tuple", errCode="xfie:ElementIsNotXbrlConcept")
    return node1.parentElement == node2.parentElement

def cu_equal(xc, p, args):
    return nodesEqual(xc, args, cu_equal_test, mustBeItems=True, nonItemErrCode="xfie:NodeIsNotXbrlItem")

def cu_equal_test(modelItem1, modelItem2):
    return c_equal_test(modelItem1, modelItem2) and u_equal_test(modelItem1, modelItem2)

def pc_equal(xc, p, args):
    return nodesEqual(xc, args, pc_equal_test, mustBeItems=True, nonItemErrCode="xfie:NodeIsNotXbrlItem")

def pc_equal_test(modelItem1, modelItem2):
    return p_equal_test(modelItem1, modelItem2) and c_equal_test(modelItem1, modelItem2)

def pcu_equal(xc, p, args):
    return nodesEqual(xc, args, pcu_equal_test, mustBeItems=True, nonItemErrCode="xfie:NodeIsNotXbrlItem")

def pcu_equal_test(modelItem1, modelItem2):
    return p_equal_test(modelItem1, modelItem2) and c_equal_test(modelItem1, modelItem2) and u_equal_test(modelItem1, modelItem2)

def start_equal(xc, p, args):
    return date_equal_test(xc, p, args, False)

def end_equal(xc, p, args):
    return date_equal_test(xc, p, args, True)

def taxonomy_refs(xc, p, args):
    return [ref.referringModelObject.xAttributes.get("{http://www.w3.org/1999/xlink}href").xValue # need typed value
            for ref in sorted(xc.modelXbrl.modelDocument.referencesDocument.values(),
                              key=lambda r:r.referringModelObject.objectIndex)
            if ref.referringModelObject.localName == "schemaRef"]

def date_equal_test(xc, p, args, instantEndDate):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    date1 = atomicArg(xc, p, args, 0, "xbrldi:dateUnion", missingArgFallback=(), emptyFallback=None)
    if not isinstance(date1, (DateTime,datetime.date)):
        raise XPathContext.FunctionArgType(1,"xbrldi:dateUnion")
    date2 = atomicArg(xc, p, args, 1, "xbrldi:dateUnion", missingArgFallback=(), emptyFallback=None)
    if not isinstance(date1, (DateTime,datetime.date)):
        raise XPathContext.FunctionArgType(2,"xbrldi:dateUnion")
    return dateUnionEqual(date1, date2, instantEndDate)

def nodes_correspond(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    node1 = nodeArg(xc, args, 0, "node()?", missingArgFallback=(), emptyFallback=())
    node2 = nodeArg(xc, args, 1, "node()?", missingArgFallback=(), emptyFallback=())
    if node1 == ():
        if node2 == (): return True
        return False
    if node2 == (): return False
    return XbrlUtil.nodesCorrespond(xc.modelXbrl, node1, node2, xc.modelXbrl)

def facts_in_instance(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args)
    return inst.factsInInstance

def items_in_instance(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args)
    return [f for f in inst.factsInInstance if f.isItem]

def tuples_in_instance(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args)
    return [f for f in inst.factsInInstance if f.isTuple]

def items_in_tuple(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    parentTuple = xbrlTuple(xc, args, 0)
    return [f for f in parentTuple.modelTupleFacts if f.isItem]

def tuples_in_tuple(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    parentTuple = xbrlTuple(xc, args, 0)
    return [f for f in parentTuple.modelTupleFacts if f.isTuple]

def non_nil_facts_in_instance(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args)
    return [f for f in inst.factsInInstance if (f.isItem or f.isTuple) and not f.isNil]

def concept(xc, p, args):
    qnConcept = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    srcConcept = xc.modelXbrl.qnameConcepts.get(qnConcept)
    if srcConcept is None or not (srcConcept.isItem or srcConcept.isTuple) or srcConcept.qname is None or srcConcept.qname.namespaceURI == XbrlConst.xbrli:
        raise XPathContext.XPathException(p, 'xfie:invalidConceptQName', _('Argument 1 {0} is not a concept in the DTS.').format(qnConcept))
    return srcConcept

def concept_balance(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    balance = concept(xc,p,args).get("{http://www.xbrl.org/2003/instance}balance")
    if balance is None:
        balance = ""
    return balance

def concept_period_type(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    return concept(xc,p,args).get("{http://www.xbrl.org/2003/instance}periodType")

def concept_custom_attribute(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    qnAttr = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)
    if qnAttr is None: raise XPathContext.FunctionArgType(2,"xs:QName")
    element = concept(xc,p,args)
    return element_attribute(element, qnAttr)

def concept_data_type(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    typeQname = concept(xc,p,args).typeQname
    if typeQname is None or typeQname.localName.endswith(anonymousTypeSuffix):
        return ()
    return typeQname

def concept_data_type_derived_from(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    qnType = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)
    if qnType is None: raise XPathContext.FunctionArgType(2,"xs:QName")
    return concept(xc,p,args).instanceOfType(qnType)

def concept_substitutions(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    return concept(xc,p,args).substitutionGroupQnames

def concepts_from_local_name(xc, p, args):
    if not 1 <= len(args) <= 2: raise XPathContext.FunctionNumArgs()
    localName = stringArg(xc, args, 0, "xs:string")
    if len(args) == 2:
        nsPattern = re.compile(stringArg(xc, args, 1, "xs:string"))
        return [c.qname for c in xc.modelXbrl.nameConcepts.get(localName,())
                if (c.isItem or c.isTuple) and bool(nsPattern.search(c.qname.namespaceURI))]
    else:
        return [c.qname for c in xc.modelXbrl.nameConcepts.get(localName,())
                if c.isItem or c.isTuple]

def concepts_from_local_name_pattern(xc, p, args):
    if not 1 <= len(args) <= 2: raise XPathContext.FunctionNumArgs()
    localNamePattern = re.compile(stringArg(xc, args, 0, "xs:string"))
    if len(args) == 2:
        nsPattern = re.compile(stringArg(xc, args, 1, "xs:string"))
        return [c.qname for c in xc.modelXbrl.qnameConcepts.values()
                if (c.isItem or c.isTuple) and bool(localNamePattern.search(c.name)) and bool(nsPattern.search(c.qname.namespaceURI))]
    else:
        return [c.qname for c in xc.modelXbrl.qnameConcepts.values()
                if (c.isItem or c.isTuple) and bool(localNamePattern.search(c.name))]

def filter_member_network_selection(xc, p, args):
    if len(args) != 5: raise XPathContext.FunctionNumArgs()
    qnDim = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    qnMem = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)
    linkroleURI = stringArg(xc, args, 2, "xs:string")
    arcroleURI = stringArg(xc, args, 3, "xs:string")
    axis = stringArg(xc, args, 4, "xs:string")
    if not axis in ('descendant-or-self', 'child-or-self', 'descendant', 'child'):
        return ()
    dimConcept = xc.modelXbrl.qnameConcepts.get(qnDim)
    if dimConcept is None or not dimConcept.isDimensionItem:
        raise XPathContext.XPathException(p, 'xfie:invalidDimensionQName', _('Argument 1 {0} is not a dimension concept QName.').format(qnDim))
    memConcept = xc.modelXbrl.qnameConcepts.get(qnMem)
    if memConcept is None or not memConcept.isDomainMember:
        # removed error 2011-03-10: raise XPathContext.XPathException(p, 'xfie:unrecognisedExplicitDimensionValueQName', _('Argument 1 {0} is not a member concept QName.').format(qnMem))
        return ()
    relationshipSet = xc.modelXbrl.relationshipSet(arcroleURI, linkroleURI)
    if relationshipSet is not None:
        members = set()
        ''' removed 2011-03-10:
        linkQnames = set()
        arcQnames = set()
        '''
        if axis.endswith("-or-self"):
            members.add(qnMem)
        fromRels = relationshipSet.fromModelObject(memConcept)
        if fromRels is not None:
            filter_member_network_members(relationshipSet, fromRels, axis.startswith("descendant"), members=members)
            ''' removed 2011-03-10:
            if len(linkQnames) > 1 or len(arcQnames) > 1:
                raise XPathContext.XPathException(p, 'xfie:ambiguousFilterMemberNetwork',
                          _('Network of linkrole {0} and arcrole {1} dimension {2} from {3} is ambiguous because of multiple link elements, {4}, or arc elements {5}').format(
                            linkroleURI, arcroleURI, qnDim, qnMem, linkQnames, arcQnames))
            '''
            return members
        # no fromRels, must be a toRel or else the qname is not in the member network
        if relationshipSet.toModelObject(memConcept):
            return members  # valid situation, the member exists as a leaf node
    # removed error 2011-03-10: raise XPathContext.XPathException(p, 'xfie:unrecognisedExplicitDimensionValueQName', _('Argument 1 {0} member is not in the network.').format(qnMem))
    return ()

def filter_member_network_members(relationshipSet, fromRels, recurse, members=None, relationships=None, linkQnames=None, arcQnames=None):
    if members is None:
        members = set()
    for modelRel in fromRels:
        toConcept = modelRel.toModelObject
        toConceptQname = toConcept.qname
        if linkQnames is not None:
            linkQnames.add(modelRel.linkQname)
        if arcQnames is not None:
            arcQnames.add(modelRel.qname)
        if toConceptQname not in members:
            members.add(toConceptQname)
            if relationships is not None:
                relationships.add(modelRel)
            if recurse:
                filter_member_network_members(relationshipSet, relationshipSet.fromModelObject(toConcept), recurse, members, relationships, linkQnames, arcQnames)

def filter_member_DRS_selection(xc, p, args):
    if len(args) != 5: raise XPathContext.FunctionNumArgs()
    qnDim = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    qnPriItem = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)
    qnMem = qnameArg(xc, p, args, 2, 'QName', emptyFallback=None)
    linkroleURI = stringArg(xc, args, 3, "xs:string")
    if not linkroleURI:  # '' or ()
        linkroleURI = None  # select all ELRs
    axis = stringArg(xc, args, 4, "xs:string")
    if not axis in ('DRS-descendant', 'DRS-child'):
        return ()
    memSelectionQnames = set()
    dimConcept = xc.modelXbrl.qnameConcepts.get(qnDim)
    if dimConcept is None or dimConcept.qname is None or dimConcept.qname.namespaceURI == XbrlConst.xbrli:
        raise XPathContext.XPathException(p, 'xfie:invalidDimensionQName', _('Argument 1 {0} is not in the DTS.').format(qnDim))
    elif not dimConcept.isDimensionItem:
        raise XPathContext.XPathException(p, 'xfie:invalidDimensionQName', _('Argument 1 {0} is not a dimension.').format(qnDim))
    priItemConcept = xc.modelXbrl.qnameConcepts.get(qnPriItem)
    if priItemConcept is None  or priItemConcept.qname is None  or priItemConcept.qname.namespaceURI == XbrlConst.xbrli:
        raise XPathContext.XPathException(p, 'xfie:invalidPrimaryItemConceptQName', _('Argument 2 {0} is not in the DTS.').format(qnPriItem))
    elif not priItemConcept.isPrimaryItem:
        raise XPathContext.XPathException(p, 'xfie:invalidPrimaryItemConceptQName', _('Argument 2 {0} is not a primary item.').format(qnPriItem))
    memConcept = xc.modelXbrl.qnameConcepts.get(qnMem)
    if memConcept is None or not memConcept.isDomainMember or not dimConcept.isDimensionItem:
        # not an error, just don't find anything
        return ()
    for hcELR, hcRels in priItemElrHcRels(xc, priItemConcept, linkroleURI).items():
        if not linkroleURI or linkroleURI == hcELR:
            for hasHcRel in hcRels:
                hcConcept = hasHcRel.toModelObject
                if hasHcRel.arcrole == XbrlConst.all:
                    dimELR = (hasHcRel.targetRole or hcELR)
                    for hcDimRel in xc.modelXbrl.relationshipSet(XbrlConst.hypercubeDimension, dimELR).fromModelObject(hcConcept):
                        if dimConcept == hcDimRel.toModelObject:
                            filter_member_DRS_members(xc,
                                                      xc.modelXbrl.relationshipSet(XbrlConst.dimensionDomain,
                                                                                   (hcDimRel.targetRole or dimELR))
                                                      .fromModelObject(dimConcept),
                                                      axis,
                                                      memConcept,
                                                      False,
                                                      set(),
                                                      memSelectionQnames)
    return memSelectionQnames

def filter_member_DRS_members(xc, fromRels, axis, memConcept, inSelection, visited, memSelectionQnames):
    for rel in fromRels:
        toConcept = rel.toModelObject
        toConceptQname = toConcept.qname
        nestedSelection = inSelection
        if rel.fromModelObject == memConcept or inSelection:  # from is the asked-for parent
            memSelectionQnames.add(toConceptQname) # to is a child or descendant
            nestedSelection = True
        if toConceptQname not in visited and (not nestedSelection or axis == "DRS-descendant"):
            visited.add(toConcept)
            filter_member_DRS_members(xc,
                                      xc.modelXbrl.relationshipSet(XbrlConst.domainMember,
                                                                   (rel.targetRole or rel.linkrole))
                                      .fromModelObject(toConcept),
                                      axis,
                                      memConcept,
                                      nestedSelection,
                                      visited,
                                      memSelectionQnames)
            visited.discard(toConcept)

def dimension_default(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    qnDim = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    dimConcept = xc.modelXbrl.qnameConcepts.get(qnDim)
    if dimConcept is None or dimConcept.qname is None or dimConcept.qname.namespaceURI == XbrlConst.xbrli:
        raise XPathContext.XPathException(p, 'xfie:invalidDimensionQName', _('Argument 1 {0} is not in the DTS.').format(qnDim))
    elif not dimConcept.isDimensionItem:
        raise XPathContext.XPathException(p, 'xfie:invalidDimensionQName', _('Argument 1 {0} is not a dimension.').format(qnDim))
    for dimDefRel in xc.modelXbrl.relationshipSet(XbrlConst.dimensionDefault).fromModelObject(dimConcept):
        dimConcept = dimDefRel.toModelObject
        if dimConcept is not None and dimConcept.isDomainMember:
            return [dimConcept.qname]
    return []

def fact_segment_remainder(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        return context.segNonDimValues
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def fact_scenario_remainder(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        return context.scenNonDimValues
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def fact_dim_value(xc, p, args, dimType):
    context = item_context(xc, args)
    qnDim = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)
    dimConcept = xc.modelXbrl.qnameConcepts.get(qnDim)
    if dimConcept is None or not dimConcept.isDimensionItem:
        raise XPathContext.XPathException(p,
                                          'xfie:invalid{0}DimensionQName'.format(dimType),
                                          _('Argument 1 {0} is not a dimension concept QName.').format(qnDim))
    if context is not None:
        return context.dimValue(qnDim)
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def fact_has_explicit_dimension(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    dimValue = fact_dim_value(xc, p, args, "Explicit")
    return dimValue is not None and (isinstance(dimValue,QName) or
                                     dimValue.isExplicit)

def fact_has_typed_dimension(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    dimValue = fact_dim_value(xc, p, args, "Typed")
    return dimValue is not None and not isinstance(dimValue,QName) and dimValue.isTyped

def fact_explicit_dimension_value_value(xc, p, args):
    context = item_context(xc, args)
    if context is not None:
        qn = qnameArg(xc, p, args, 1, 'QName', emptyFallback=())
        if qn == (): raise XPathContext.FunctionArgType(2,"xbrl:QName")
        dimConcept = xc.modelXbrl.qnameConcepts.get(qn) # check qname is explicit dimension
        if dimConcept is None or not dimConcept.isExplicitDimension:
            raise XPathContext.XPathException(p, 'xfie:invalidExplicitDimensionQName', _('dimension does not specify an explicit dimension'))
        dimValue = context.dimValue(qn)
        if isinstance(dimValue, ModelDimensionValue) and dimValue.isExplicit:
            return dimValue.memberQname # known to be valid given instance is valid
        elif isinstance(dimValue, QName): #default, check if this is valid
            ''' removed 2011-03-01 FWG clarification that default always applies
            #note that there's no way to check one dimension without full set of others for validity
            modelItem = xc.modelItem(args[0][0])
            itemConcept = modelItem.concept
            from arelle.ValidateXbrlDimensions import checkPriItemDimValueValidity
            memConcept = xc.modelXbrl.qnameConcepts.get(dimValue)
            # remove check for pri item validity per FWG meeting notes 2011-01-13
            if itemConcept: # and checkPriItemDimValueValidity(xc, itemConcept, dimConcept, memConcept):
                return dimValue
            '''
            return dimValue
        return () # not an applicable primary item for default dimension
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def fact_has_explicit_dimension_value(xc, p, args):
    if len(args) != 3: raise XPathContext.FunctionNumArgs()
    return qnameArg(xc, p, args, 2, 'QName', emptyFallback=()) == fact_explicit_dimension_value_value(xc, p, args)

def fact_explicit_dimension_value(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    return fact_explicit_dimension_value_value(xc, p, args)

def fact_typed_dimension_value(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        qn = qnameArg(xc, p, args, 1, 'QName', emptyFallback=())
        if qn == (): raise XPathContext.FunctionArgType(2,"xbrl:QName")
        modelConcept = xc.modelXbrl.qnameConcepts.get(qn) # check qname is explicit dimension
        if modelConcept is None or not modelConcept.isTypedDimension:
            raise XPathContext.XPathException(p, 'xfie:invalidTypedDimensionQName', _('dimension does not specify a typed dimension'))
        result = context.dimValue(qn)
        return result if result is not None else ()
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def fact_typed_dimension_simple_value(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        qn = qnameArg(xc, p, args, 1, 'QName', emptyFallback=())
        if qn == (): raise XPathContext.FunctionArgType(2,"xbrl:QName")
        modelConcept = xc.modelXbrl.qnameConcepts.get(qn) # check qname is explicit dimension
        if modelConcept is None or not modelConcept.isTypedDimension:
            raise XPathContext.XPathException(p, 'xfie:invalidTypedDimensionQName', _('dimension does not specify a typed dimension'))
        result = context.dimValue(qn)
        if result is None:
            return ()
        typedMember = result.typedMember
        if typedMember.get("{http://www.w3.org/2001/XMLSchema-instance}nil", "false") in ("true","1"):
            return ()
        if typedMember.xValid == VALID_NO_CONTENT:
            for _child in typedMember.iterchildren(): # has children
                raise XPathContext.XPathException(p, 'xqt-err:FOTY0012', _('dimension domain element is not atomizable'))
            return () # no error if no children
        return typedMember.xValue
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def fact_explicit_dimensions(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        return set(qn for qn, dim in context.qnameDims.items() if dim.isExplicit) | xc.modelXbrl.qnameDimensionDefaults.keys()
    return set()

def fact_typed_dimensions(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    context = item_context(xc, args)
    if context is not None:
        return set(qn for qn, dim in context.qnameDims.items() if dim.isTyped)
    return set()

def fact_dimension_s_equal2(xc, p, args):
    if len(args) != 3: raise XPathContext.FunctionNumArgs()
    context1 = item_context(xc, args, i=0)
    context2 = item_context(xc, args, i=1)
    if context1 is not None:
        if context2 is not None:
            qn = qnameArg(xc, p, args, 2, 'QName', emptyFallback=())
            if qn == (): raise XPathContext.FunctionArgType(3,"xbrl:QName")
            modelConcept = xc.modelXbrl.qnameConcepts.get(qn) # check qname is explicit dimension
            if modelConcept is None or not modelConcept.isDimensionItem:
                # raise XPathContext.XPathException(p, 'xfie:invalidTypedDimensionQName', _('dimension does not specify a typed dimension'))
                return False
            dimValue1 = context1.dimValue(qn)
            dimValue2 = context2.dimValue(qn)
            if dimValue1 is not None and isinstance(dimValue1,ModelDimensionValue):
                return dimValue1.isEqualTo(dimValue2, equalMode=XbrlUtil.S_EQUAL2)
            elif dimValue2 is not None and isinstance(dimValue2,ModelDimensionValue):
                return dimValue2.isEqualTo(dimValue1, equalMode=XbrlUtil.S_EQUAL2)
            return dimValue1 == dimValue2
        raise XPathContext.FunctionArgType(2,"xbrl:item")
    raise XPathContext.FunctionArgType(1,"xbrl:item")

def linkbase_link_roles(xc, p, args):
    if len(args) > 2: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 1)
    arcroleURI = stringArg(xc, args, 0, "xs:string")
    relationshipSet = inst.relationshipSet(arcroleURI)
    if relationshipSet:
        return [anyURI(linkrole) for linkrole in relationshipSet.linkRoleUris]
    return ()

def navigate_relationships(xc, p, args):
    raise xfiFunctionNotAvailable()

def concept_label(xc, p, args):
    if not 4 <= len(args) <= 5: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 4)
    qnSource = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    srcConcept = inst.qnameConcepts.get(qnSource)
    if srcConcept is None:
        return ""
    linkroleURI = stringArg(xc, args, 1, "xs:string", emptyFallback='')
    if not linkroleURI: linkroleURI = XbrlConst.defaultLinkRole
    labelroleURI = stringArg(xc, args, 2, "xs:string", emptyFallback='')
    if not labelroleURI: labelroleURI = XbrlConst.standardLabel
    lang = stringArg(xc, args, 3, "xs:string", emptyFallback='')
    relationshipSet = inst.relationshipSet(XbrlConst.conceptLabel,linkroleURI)
    if relationshipSet is not None:
        label = relationshipSet.label(srcConcept, labelroleURI, lang)
        if label is not None: return label
    return ""


def arcrole_definition(xc, p, args):
    if len(args) > 2: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 1)
    arcroleURI = stringArg(xc, args, 0, "xs:string", emptyFallback='')
    modelArcroleTypes = inst.arcroleTypes.get(arcroleURI)
    if modelArcroleTypes is not None and len(modelArcroleTypes) > 0:
        arcroledefinition = modelArcroleTypes[0].definition
        if arcroledefinition is not None: return arcroledefinition
    return ()

def role_definition(xc, p, args):
    if len(args) > 2: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 1)
    roleURI = stringArg(xc, args, 0, "xs:string", emptyFallback='')
    modelRoleTypes = inst.roleTypes.get(roleURI)
    if modelRoleTypes is not None and len(modelRoleTypes) > 0:
        roledefinition = modelRoleTypes[0].definition
        if roledefinition is not None: return roledefinition
    return ()

def fact_footnotes(xc, p, args):
    if len(args) > 6: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 5)
    itemObj = item(xc, args)
    linkroleURI = stringArg(xc, args, 1, "xs:string", emptyFallback='')
    if not linkroleURI: linkroleURI = XbrlConst.defaultLinkRole
    arcroleURI = stringArg(xc, args, 2, "xs:string", emptyFallback='')
    if not arcroleURI: arcroleURI = XbrlConst.factFootnote
    footnoteroleURI = stringArg(xc, args, 3, "xs:string", emptyFallback='')
    if not footnoteroleURI: footnoteroleURI = XbrlConst.footnote
    lang = stringArg(xc, args, 4, "xs:string", emptyFallback='')
    relationshipSet = inst.relationshipSet(arcroleURI,linkroleURI)
    if relationshipSet: # must return empty sequence, not None if no footnotes match filters
        return relationshipSet.label(itemObj, footnoteroleURI, lang, returnMultiple=True) or ()
    return ()

def concept_relationships(xc, p, args, nestResults=False):
    lenArgs = len(args)
    if not 4 <= lenArgs <= 8: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 7)
    qnSource = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    linkroleURI = stringArg(xc, args, 1, "xs:string")
    if not linkroleURI:
        linkroleURI = XbrlConst.defaultLinkRole
    elif linkroleURI == "XBRL-all-linkroles":
        linkroleURI = None
    arcroleURI = stringArg(xc, args, 2, "xs:string")
    axis = stringArg(xc, args, 3, "xs:string")
    if not axis in ('descendant', 'child', 'ancestor', 'parent', 'sibling', 'sibling-or-self'):
        raise XPathContext.FunctionArgType(3, "'descendant', 'child', 'ancestor', 'parent', 'sibling' or 'sibling-or-self'",
                                           errCode="xfie:InvalidConceptRelationParameters")
    if qnSource != XbrlConst.qnXfiRoot:
        srcConcept = inst.qnameConcepts.get(qnSource)
        if srcConcept is None:
            return ()
    if lenArgs > 4:
        generations = numericArg(xc, p, args, 4, "xs:integer", convertFallback=0)
        if axis in ('child', 'parent', 'sibling', 'sibling-or-self') and generations != 1:
            raise XPathContext.FunctionArgType(4, "generations must be 1 for 'child', 'parent', 'sibling' or 'sibling-or-self' axis",
                                               errCode="xfie:InvalidConceptRelationParameters")
    elif axis in ('child', 'parent', 'sibling', 'sibling-or-self'):
        generations = 1
    else:
        generations = 0
    if axis in ('child','parent','sibling', 'sibling-or-self'):
        generations = 1
    if axis == 'child':
        axis = 'descendant'
    elif axis == 'parent':
        axis = 'ancestor'
    if lenArgs > 5:
        qnLink = qnameArg(xc, p, args, 5, 'QName', emptyFallback=None)
    else:
        qnLink = None
    if lenArgs > 5:
        qnArc = qnameArg(xc, p, args, 6, 'QName', emptyFallback=None)
    else:
        qnArc = None

    removeSelf = axis == 'sibling'
    relationshipSet = inst.relationshipSet(arcroleURI, linkroleURI, qnLink, qnArc)
    if relationshipSet:
        result = []
        visited = {qnSource}
        if qnSource == XbrlConst.qnXfiRoot:
            if axis in ('sibling', 'sibling-or-self', 'ancestor'):
                return []
            roots = relationshipSet.rootConcepts
            visited = {c.qname for c in roots}
            rels = [rel for c in roots for rel in relationshipSet.fromModelObject(c)]
            if generations == 1:
                return rels
            if generations > 1:
                generations -= 1
        elif axis == 'descendant':
            rels = relationshipSet.fromModelObject(srcConcept)
        elif axis == 'ancestor': # includes first pass on parents of object to get sibling
            rels = relationshipSet.toModelObject(srcConcept)
        elif axis in ('sibling', 'sibling-or-self'):
            rels = relationshipSet.toModelObject(srcConcept)
            if rels:
                rels = relationshipSet.fromModelObject(rels[0].fromModelObject)
                axis = 'descendant'
            else: # must be a root, never has any siblings
                return []
        if rels:
            concept_relationships_step(xc, inst, relationshipSet, rels, axis, generations, result, visited, nestResults)
            if removeSelf:
                for i, rel in enumerate(result):
                    if rel.toModelObject == srcConcept:
                        result.pop(i)
                        break
            return result
    return ()

def concept_relationships_step(xc, inst, relationshipSet, rels, axis, generations, result, visited, nestResults):
    if rels:
        for modelRel in rels:
            concept = modelRel.toModelObject if axis == 'descendant' else modelRel.fromModelObject
            conceptQname = concept.qname
            result.append(modelRel)
            if generations > 1 or (generations == 0 and conceptQname not in visited):
                nextGen = (generations - 1) if generations > 1 else 0
                if generations == 0:
                    visited.add(conceptQname)
                if axis == 'descendant':
                    if relationshipSet.arcrole == "XBRL-dimensions":
                        stepRelationshipSet = inst.relationshipSet("XBRL-dimensions", modelRel.consecutiveLinkrole)
                    else:
                        stepRelationshipSet = relationshipSet
                    stepRels = stepRelationshipSet.fromModelObject(concept)
                else:
                    if relationshipSet.arcrole == "XBRL-dimensions":
                        stepRelationshipSet = inst.relationshipSet("XBRL-dimensions")
                        # search all incoming relationships for those with right consecutiveLinkrole
                        stepRels = [rel
                                    for rel in stepRelationshipSet.toModelObject(concept)
                                    if rel.consectuiveLinkrole == modelRel.linkrole]
                    else:
                        stepRelationshipSet = relationshipSet
                        stepRels = stepRelationshipSet.toModelObject(concept)
                if nestResults: # nested results are in a sub-list
                    nestedList = []
                else: # nested results flattened in top level results
                    nestedList = result
                concept_relationships_step(xc, inst, stepRelationshipSet, stepRels, axis, nextGen, nestedList, visited, nestResults)
                if nestResults and nestedList:  # don't append empty nested results
                    result.append(nestedList)
                if generations == 0:
                    visited.discard(conceptQname)

def relationship_from_concept(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    modelRel = anytypeArg(xc, args, 0, "arelle:ModelRelationship", None)
    if modelRel is not None:
        return modelRel.fromModelObject.qname
    raise XPathContext.FunctionArgType(1,"arelle:modelRelationship")

def relationship_to_concept(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    modelRel = anytypeArg(xc, args, 0, "arelle:ModelRelationship", None)
    if modelRel is not None:
        return modelRel.toModelObject.qname
    raise XPathContext.FunctionArgType(1,"arelle:modelRelationship")

def distinct_nonAbstract_parent_concepts(xc, p, args):
    lenArgs = len(args)
    if not 2 <= lenArgs <= 3: raise XPathContext.FunctionNumArgs()
    inst = instance(xc, p, args, 2)
    linkroleURI = stringArg(xc, args, 0, "xs:string")
    if not linkroleURI:
        linkroleURI = XbrlConst.defaultLinkRole
    arcroleURI = stringArg(xc, args, 1, "xs:string")
    # TBD allow instance as arg 2

    result = set()
    relationshipSet = inst.relationshipSet(arcroleURI, linkroleURI)
    if relationshipSet:
        for rel in relationshipSet.modelRelationships:
            fromModelObject = rel.fromModelObject
            toModelObject = rel.toModelObject
            if (isinstance(fromModelObject, ModelConcept) and
                isinstance(toModelObject, ModelConcept) and
                not fromModelObject.isAbstract and
                not toModelObject.isAbstract):
                result.add(fromModelObject.qname)
    return result

def relationship_element_attribute(xc, p, args, elementParent=False):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    modelRel = anytypeArg(xc, args, 0, "arelle:ModelRelationship", None)
    if modelRel is None: raise XPathContext.FunctionArgType(1,"arelle:modelRelationship")
    qnAttr = qnameArg(xc, p, args, 1, 'QName', emptyFallback=None)
    if qnAttr is None: raise XPathContext.FunctionArgType(2,"xs:QName")
    element = modelRel.arcElement
    if elementParent: element = element.getparent()
    return element_attribute(element, qnAttr)

def element_attribute(element, attrQname):
    attrTag = attrQname.clarkNotation
    modelAttribute = None
    try:
        modelAttribute = element.xAttributes[attrTag]
    except (AttributeError, TypeError, IndexError, KeyError):
        # may be lax or deferred validated
        try:
            xmlValidate(element.modelXbrl, element, attrQname)
            modelAttribute = element.xAttributes[attrTag]
        except (AttributeError, TypeError, IndexError, KeyError):
            pass
    if modelAttribute is None:
        value = element.get(attrTag)
        if value is not None:
            return value
    elif modelAttribute.xValid >= VALID:
        return modelAttribute.xValue
    return ()

def relationship_attribute(xc, p, args):
    return relationship_element_attribute(xc, p, args)

def relationship_link_attribute(xc, p, args):
    return relationship_element_attribute(xc, p, args, elementParent=True)

def element_name(xc, p, args, elementParent=False):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    modelRel = anytypeArg(xc, args, 0, "arelle:ModelRelationship", None)
    if modelRel is None: raise XPathContext.FunctionArgType(1,"arelle:modelRelationship")
    element = modelRel.arcElement
    if elementParent: element = element.getparent()
    return qname(element)

def relationship_name(xc, p, args):
    return element_name(xc, p, args)

def relationship_link_name(xc, p, args):
    return element_name(xc, p, args, elementParent=True)

def xbrl_instance(xc, p, args):
    raise xfiFunctionNotAvailable()

def  format_number(xc, p, args):
    if len(args) != 2: raise XPathContext.FunctionNumArgs()
    value = numericArg(xc, p, args, 0, missingArgFallback='NaN', emptyFallback='NaN')
    picture = stringArg(xc, args, 1, "xs:string", missingArgFallback='', emptyFallback='')
    try:
        return format_picture(xc.modelXbrl.locale, value, picture)
    except ValueError as err:
        raise XPathContext.XPathException(p, 'xfie:invalidPictureSyntax', str(err) )

# note that this function was initially in plugin functionsXmlCreation when it was named xfxc:element
def  create_element(xc, p, args):
    if not 2 <= len(args) <= 4: raise XPathContext.FunctionNumArgs()
    qn = qnameArg(xc, p, args, 0, 'QName', emptyFallback=None)
    attrArg = flattenSequence(args[1])
    # attributes have to be pairs
    if attrArg:
        if (len(attrArg) & 1 or
            any((not isinstance(arg, (QName, str))) or
                (isinstance(arg,str) and NCNamePattern.match(arg) is None)
                for i in range(0, len(attrArg),2)
                for arg in (attrArg[i],))):
            raise XPathContext.FunctionArgType(1,"((xs:qname|xs:string),xs:anyAtomicValue)", errCode="xfie:AttributesNotNameValuePairs")
        else:
            attrParam = [(attrArg[i],attrArg[i+1]) # need name-value pairs for XmlUtil function
                         for i in range(0, len(attrArg),2)]
    else:
        attrParam = None

    value = atomicArg(xc, p, args, 2, "xs:anyAtomicType", emptyFallback='')
    if not value: # be sure '' is None so no text node is created
        value = None
    if len(args) < 4:
        childElements = None
    else:
        childElements = xc.flattenSequence(args[3])
    if value and childElements:
        raise XPathContext.FunctionArgType(1,str(value), errCode="xfie:MixedContentError")

    # scratchpad instance document emulates fn:doc( ) to hold XML nodes
    scratchpadXmlDocUrl = "http://www.xbrl.org/2012/function/creation/xml_scratchpad.xml"
    if scratchpadXmlDocUrl in xc.modelXbrl.urlDocs:
        modelDocument = xc.modelXbrl.urlDocs[scratchpadXmlDocUrl]
    else:
        # create scratchpad xml document
        # this will get the fake instance document in the list of modelXbrl docs so that it is garbage collected
        from arelle import ModelDocument
        modelDocument = ModelDocument.create(xc.modelXbrl,
                                             ModelDocument.Type.UnknownXML,
                                             scratchpadXmlDocUrl,
                                             initialXml="<xfc:dummy xmlns:xfc='http://www.xbrl.org/2012/function/creation'/>")

    newElement = XmlUtil.addChild(modelDocument.xmlRootElement,
                                  qn,
                                  attributes=attrParam,
                                  text=value)
    if childElements:
        for element in childElements:
            if isinstance(element, etree.ElementBase):
                newElement.append(element)

    # node myst be validated for use in instance creation (typed dimension references)
    xmlValidate(xc.modelXbrl, newElement)

    return newElement

def any_identifier(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    for cntx in xc.modelXbrl.contextsInUse:
        return cntx.entityIdentifierElement
    return ()

def unique_identifiers(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    distinctIdentifiers = {}
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.entityIdentifier not in distinctIdentifiers:
            distinctIdentifiers[cntx.entityIdentifier] = cntx.entityIdentifierElement
    return [e for k,e in sorted(distinctIdentifiers.items(), key=lambda i:i[0])]

def single_unique_identifier(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    return len(set(cntx.entityIdentifier for cntx in xc.modelXbrl.contextsInUse)) == 1

def any_start_date(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.isStartEndPeriod:
            return cntx.startDatetime
    return ()

def unique_start_dates(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    distinctStartDates = set()
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.isStartEndPeriod:
            distinctStartDates.add(cntx.startDatetime)
    return [sorted(distinctStartDates, key=lambda d:(d.tzinfo is None,d))]

def single_unique_start_date(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    return len(set(cntx.startDatetime for cntx in xc.modelXbrl.contextsInUse if cntx.isStartEndPeriod)) == 1

def any_end_date(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.isStartEndPeriod:
            return cntx.endDatetime
    return ()

def unique_end_dates(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    distinctStartDates = set()
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.isStartEndPeriod:
            distinctStartDates.add(cntx.endDatetime)
    return [sorted(distinctStartDates, key=lambda d:(d.tzinfo is None,d))]

def single_unique_end_date(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    return len(set(cntx.endDatetime for cntx in xc.modelXbrl.contextsInUse if cntx.isStartEndPeriod)) == 1

def any_instant_date(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.isInstantPeriod:
            return cntx.instantDatetime
    return ()

def unique_instant_dates(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    distinctStartDates = set()
    for cntx in xc.modelXbrl.contextsInUse:
        if cntx.isInstantPeriod:
            distinctStartDates.add(cntx.instantDatetime)
    return [sorted(distinctStartDates, key=lambda d:(d.tzinfo is None,d))]

def single_unique_instant_date(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    return len(set(cntx.instantDatetime for cntx in xc.modelXbrl.contextsInUse if cntx.isInstantPeriod)) == 1

def filingIndicatorValues(inst, filedValue):
    filingIndicators = set()
    for fact in inst.factsByQname[XbrlConst.qnEuFiIndFact]:
        if fact.parentElement.qname == XbrlConst.qnEuFiTuple and fact.get(XbrlConst.cnEuFiIndAttr,"true") == filedValue:
            filingIndicators.add(fact.stringValue.strip())
    for fact in inst.factsByQname[XbrlConst.qnFiFact]:
        if fact.context is not None and XbrlConst.qnFiDim in fact.context.qnameDims and fact.value.strip() == filedValue:
            fiValue = fact.context.qnameDims[XbrlConst.qnFiDim].stringValue.strip()
            if fiValue:
                filingIndicators.add(fiValue)
    return filingIndicators

def positive_filing_indicators(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    return sorted(filingIndicatorValues(xc.modelXbrl, "true"))

def positive_filing_indicator(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    ind = anytypeArg(xc, args, 0, "xs:string", None)
    if ind is None: raise XPathContext.FunctionArgType(1,"xs:string")
    return ind in filingIndicatorValues(xc.modelXbrl, "true")

def negative_filing_indicators(xc, p, args):
    if len(args) != 0: raise XPathContext.FunctionNumArgs()
    return sorted(filingIndicatorValues(xc.modelXbrl, "false"))

def negative_filing_indicator(xc, p, args):
    if len(args) != 1: raise XPathContext.FunctionNumArgs()
    ind = anytypeArg(xc, args, 0, "xs:string", None)
    if ind is None: raise XPathContext.FunctionArgType(1,"xs:string")
    return ind in filingIndicatorValues(xc.modelXbrl, "false")

xfiFunctions = {
    'context': context,
    'unit': unit,
    'unit-numerator': unit_numerator,
    'unit-denominator': unit_denominator,
    'measure-name': measure_name,
    'period': period,
    'context-period': context_period,
    'is-start-end-period': is_start_end_period,
    'is-forever-period': is_forever_period,
    'is-duration-period': is_duration_period,
    'is-instant-period': is_instant_period,
    'period-start': period_start,
    'period-end': period_end,
    'period-instant': period_instant,
    'entity' : entity,
    'context-entity' : context_entity,
    'identifier': identifier,
    'context-identifier': context_identifier,
    'entity-identifier': entity_identifier,
    'identifier-value': identifier_value,
    'identifier-scheme': identifier_scheme,
    'segment': segment,
    'entity-segment': entity_segment,
    'context-segment': context_segment,
    'scenario': scenario,
    'context-scenario': context_scenario,
    'fact-identifier-value': fact_identifier_value,
    'fact-identifier-scheme': fact_identifier_scheme,
    'is-non-numeric' : non_numeric,
    'is-numeric' : numeric,
    'is-fraction' : fraction,
    'precision': precision,
    'decimals': decimals,
    'uncovered-aspect' : uncovered_aspect,
    'has-fallback-value' : has_fallback_value,
    'uncovered-non-dimensional-aspects' : uncovered_non_dimensional_aspects,
    'uncovered-dimensional-aspects': uncovered_dimensional_aspects,
    'identical-nodes': identical_nodes,
    's-equal': s_equal,
    'u-equal': u_equal,
    'v-equal': v_equal,
    'c-equal': c_equal,
    'identical-node-set' : identical_node_set,
    's-equal-set': s_equal_set,
    'v-equal-set': v_equal_set,
    'c-equal-set': c_equal_set,
    'u-equal-set': u_equal_set,
    'x-equal': x_equal,
    'duplicate-item': duplicate_item,
    'duplicate-tuple': duplicate_tuple,
    'p-equal': p_equal,
    'cu-equal': cu_equal,
    'pc-equal': pc_equal,
    'pcu-equal': pcu_equal,
    'start-equal': start_equal,
    'end-equal': end_equal,
    'taxonomy-refs': taxonomy_refs,
    'nodes-correspond': nodes_correspond,
    'facts-in-instance': facts_in_instance,
    'items-in-instance': items_in_instance,
    'tuples-in-instance': tuples_in_instance,
    'items-in-tuple': items_in_tuple,
    'tuples-in-tuple': tuples_in_tuple,
    'non-nil-facts-in-instance': non_nil_facts_in_instance,
    'concept-balance': concept_balance,
    'concept-period-type': concept_period_type,
    'concept-custom-attribute': concept_custom_attribute,
    'concept-data-type': concept_data_type,
    'concept-data-type-derived-from' : concept_data_type_derived_from,
    'concept-substitutions': concept_substitutions,
    'concepts-from-local-name': concepts_from_local_name,
    'concepts-from-local-name-pattern': concepts_from_local_name_pattern,
    'filter-member-network-selection' : filter_member_network_selection,
    'filter-member-DRS-selection' : filter_member_DRS_selection,
    'dimension-default': dimension_default,
    'fact-segment-remainder': fact_segment_remainder,
    'fact-scenario-remainder': fact_scenario_remainder,
    'fact-has-explicit-dimension': fact_has_explicit_dimension,
    'fact-has-typed-dimension': fact_has_typed_dimension,
    'fact-has-explicit-dimension-value': fact_has_explicit_dimension_value,
    'fact-explicit-dimension-value': fact_explicit_dimension_value,
    'fact-typed-dimension-value': fact_typed_dimension_value,
    'fact-typed-dimension-simple-value': fact_typed_dimension_simple_value,
    'fact-explicit-dimensions': fact_explicit_dimensions,
    'fact-typed-dimensions': fact_typed_dimensions,
    'fact-dimension-s-equal2': fact_dimension_s_equal2,
    'linkbase-link-roles': linkbase_link_roles,
    'concept-label': concept_label,
    'arcrole-definition': arcrole_definition,
    'role-definition': role_definition,
    'fact-footnotes': fact_footnotes,
    'concept-relationships': concept_relationships,
    'relationship-from-concept': relationship_from_concept,
    'relationship-to-concept': relationship_to_concept,
    'distinct-nonAbstract-parent-concepts': distinct_nonAbstract_parent_concepts,
    'relationship-attribute': relationship_attribute,
    'relationship-link-attribute': relationship_link_attribute,
    'relationship-name': relationship_name,
    'relationship-link-name': relationship_link_name,
    'xbrl-instance': xbrl_instance,
    'format-number':  format_number,
    'create-element': create_element,
    'any-identifier': any_identifier,
    'unique-identifiers': unique_identifiers,
    'single-unique-identifier': single_unique_identifier,
    'any-start-date': any_start_date,
    'unique-start-dates': unique_start_dates,
    'single-unique-start-date': single_unique_start_date,
    'any-end-date': any_end_date,
    'unique-end-dates': unique_end_dates,
    'single-unique-end-date': single_unique_end_date,
    'any-instant-date': any_instant_date,
    'unique-instant-dates': unique_instant_dates,
    'single-unique-instant-date': single_unique_instant_date,
    'positive-filing-indicators': positive_filing_indicators,
    'positive-filing-indicator': positive_filing_indicator,
    'negative-filing-indicators': negative_filing_indicators,
    'negative-filing-indicator': negative_filing_indicator,
     }



'''
Formula math functions plugin.

See COPYRIGHT.md for copyright information.
'''
from __future__ import annotations

import math
from collections.abc import Callable

from arelle.formula import XPathContext
from arelle.FunctionUtil import numericArg
from arelle.ModelValue import QName, qname
from arelle.Version import authorLabel, copyrightLabel
from arelle.formula.XPathParser import OperationDef
from arelle.typing import EmptyTuple

INF = float('inf')
MINUSINF = float('-inf')
NaN = float('nan')


def xfm_pi(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 0:
        raise XPathContext.FunctionNumArgs()
    return math.pi


def xfm_exp(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        return math.exp(x)
    return ()


def xfm_exp10(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        return math.pow(10.0, x)
    return ()


def xfm_log(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        if x == 0:
            return MINUSINF
        elif x == -1:
            return NaN
        elif x == MINUSINF:
            return NaN
        return math.log(x)
    return ()


def xfm_log10(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        if x == 0:
            return MINUSINF
        elif x == -1:
            return NaN
        elif x == MINUSINF:
            return NaN
        return math.log10(x)
    return ()


def xfm_pow(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 2:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        y = numericArg(xc, p, args, 1)  # type: ignore[no-untyped-call]
        if x == 0:
            if math.copysign(1, x) < 0:  # e.g., value is -0.0
                if y < 0:
                    # special case for odd integer exponents
                    _intY = int(y)
                    if _intY & 1 and y == _intY:  # special case for whole numbers
                        return MINUSINF
                    return INF
                elif y == 0:
                    return 1.0
                else:
                    return -0.0
            else:  # value is +0.0
                if y < 0:
                    return INF
                elif y == 0:
                    return 1.0
                else:
                    return 0.0
        try:
            return math.pow(x, y)
        except ValueError:
            return NaN  # pow(-2.5e0, 2.00000001e0) returns xs:double('NaN').
    return ()


def xfm_sqrt(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        if x == MINUSINF:
            return NaN
        elif x < 0:
            return INF
        return math.sqrt(x)
    return ()


def xfm_sin(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        if math.isinf(x):
            return NaN
        return math.sin(x)
    return ()


def xfm_cos(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        if math.isinf(x):
            return NaN
        return math.cos(x)
    return ()


def xfm_tan(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        if math.isinf(x):
            return NaN
        return math.tan(x)
    return ()


def xfm_asin(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        try:
            return math.asin(x)
        except ValueError:
            return NaN
    return ()


def xfm_acos(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        try:
            return math.acos(x)
        except ValueError:
            return NaN
    return ()


def xfm_atan(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float | EmptyTuple:
    if len(args) != 1:
        raise XPathContext.FunctionNumArgs()
    x = numericArg(xc, p, args, 0, emptyFallback=())  # type: ignore[no-untyped-call]
    if x != ():
        try:
            return math.atan(x)
        except ValueError:
            return NaN
    return ()


def xfm_atan2(
    xc: XPathContext.XPathContext,
    p: OperationDef,
    contextItem: XPathContext.ContextItem,
    args: XPathContext.ResultStack,
) -> float:
    if len(args) != 2:
        raise XPathContext.FunctionNumArgs()
    y = numericArg(xc, p, args, 0)  # type: ignore[no-untyped-call]
    x = numericArg(xc, p, args, 1)  # type: ignore[no-untyped-call]
    return math.atan2(y, x)


def xfmMathFunctions() -> dict[
    QName, Callable[[
        XPathContext.XPathContext,
        OperationDef,
        XPathContext.ContextItem,
        XPathContext.ResultStack,
    ], float | EmptyTuple]
]:
    return {
        qname("{http://www.xbrl.org/2008/function/math}xfm:pi"): xfm_pi,
        qname("{http://www.xbrl.org/2008/function/math}xfm:exp"): xfm_exp,
        qname("{http://www.xbrl.org/2008/function/math}xfm:exp10"): xfm_exp10,
        qname("{http://www.xbrl.org/2008/function/math}xfm:log"): xfm_log,
        qname("{http://www.xbrl.org/2008/function/math}xfm:log10"): xfm_log10,
        qname("{http://www.xbrl.org/2008/function/math}xfm:pow"): xfm_pow,
        qname("{http://www.xbrl.org/2008/function/math}xfm:sqrt"): xfm_sqrt,
        qname("{http://www.xbrl.org/2008/function/math}xfm:sin"): xfm_sin,
        qname("{http://www.xbrl.org/2008/function/math}xfm:cos"): xfm_cos,
        qname("{http://www.xbrl.org/2008/function/math}xfm:tan"): xfm_tan,
        qname("{http://www.xbrl.org/2008/function/math}xfm:asin"): xfm_asin,
        qname("{http://www.xbrl.org/2008/function/math}xfm:acos"): xfm_acos,
        qname("{http://www.xbrl.org/2008/function/math}xfm:atan"): xfm_atan,
        qname("{http://www.xbrl.org/2008/function/math}xfm:atan2"): xfm_atan2,
    }


__pluginInfo__ = {
    'name': 'Formula Math Functions',
    'version': '1.0',
    'description': "This plug-in adds formula math functions.  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'Formula.CustomFunctions': xfmMathFunctions,
}



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import *
try:
    from tkinter.ttk import *
except ImportError:
    from ttk import *
from arelle.CntlrWinTooltip import ToolTip
import os

class ViewTree:
    def __init__(self, modelXbrl, tabWin, tabTitle, hasToolTip=False, lang=None):
        self.tabWin = tabWin
        self.viewFrame = Frame(tabWin)
        self.viewFrame.view = self
        self.viewFrame.grid(row=0, column=0, sticky=(N, S, E, W))
        tabWin.add(self.viewFrame,text=tabTitle)
        self.tabTitle = tabTitle # for error messages
        vScrollbar = Scrollbar(self.viewFrame, orient=VERTICAL)
        hScrollbar = Scrollbar(self.viewFrame, orient=HORIZONTAL)
        self.treeView = Treeview(self.viewFrame, xscrollcommand=hScrollbar.set, yscrollcommand=vScrollbar.set)
        self.treeView.grid(row=0, column=0, sticky=(N, S, E, W))
        try:
            _isDarkTheme = bool(self.viewFrame.tk.call("tk::unsupported::MacWindowStyle", "isdark", "."))
        except TclError:
            _isDarkTheme = False
        if _isDarkTheme:
            self.treeView.tag_configure("ELR", background="#2D3033")
            self.treeView.tag_configure("even", background="#303030")
            self.treeView.tag_configure("odd", background="#212121")
        else:
            self.treeView.tag_configure("ELR", background="#E0F0FF")
            self.treeView.tag_configure("even", background="#F0F0F0")
            self.treeView.tag_configure("odd", background="#FFFFFF")
        if modelXbrl.modelManager.cntlr.isMac or modelXbrl.modelManager.cntlr.isMSW:
            highlightColor = "#%04x%04x%04x" % self.treeView.winfo_rgb("SystemHighlight")
        else:
            highlightColor = "#33339999ffff"  # using MSW value for Unix/Linux which has no named colors
        self.treeView.tag_configure("selected-ELR", background=highlightColor)
        self.treeView.tag_configure("selected-even", background=highlightColor)
        self.treeView.tag_configure("selected-odd", background=highlightColor)
        self.treeViewSelection = ()
        self.treeView.bind("<<TreeviewSelect>>", self.viewSelectionChange, '+')
        self.treeView.bind("<1>", self.onViewClick, '+')
        hScrollbar["command"] = self.treeView.xview
        hScrollbar.grid(row=1, column=0, sticky=(E,W))
        vScrollbar["command"] = self.treeView.yview
        vScrollbar.grid(row=0, column=1, sticky=(N,S))
        self.viewFrame.columnconfigure(0, weight=1)
        self.viewFrame.rowconfigure(0, weight=1)
        self.modelXbrl = modelXbrl
        self.hasToolTip = hasToolTip
        self.toolTipText = StringVar()
        if hasToolTip:
            self.treeView.bind("<Motion>", self.motion, '+')
            self.treeView.bind("<Leave>", self.leave, '+')
            self.toolTipText = StringVar()
            self.toolTip = ToolTip(self.treeView,
                                   textvariable=self.toolTipText,
                                   wraplength=480,
                                   follow_mouse=True,
                                   state="disabled")
            self.toolTipColId = None
            self.toolTipRowId = None
        self.modelXbrl = modelXbrl
        self.lang = lang
        self.labelrole = None
        self.nameIsPrefixed = False
        if modelXbrl:
            modelXbrl.views.append(self)
            if not lang:
                self.lang = modelXbrl.modelManager.defaultLang

    def clearTreeView(self):
        self.treeViewSelection = ()
        for node in self.treeView.get_children():
            self.treeView.delete(node)

    def viewSelectionChange(self, event=None):
        for node in self.treeViewSelection:
            if self.treeView.exists(node):
                priorTags = self.treeView.item(node)["tags"]
                if priorTags:
                    priorBgTag = priorTags[0]
                    if priorBgTag.startswith("selected-"):
                        self.treeView.item(node, tags=(priorBgTag[9:],))
        self.treeViewSelection = self.treeView.selection()
        for node in self.treeViewSelection:
            priorTags = self.treeView.item(node)["tags"]
            if priorTags:
                self.treeView.item(node, tags=("selected-" + priorTags[0],))

    def onViewClick(self, *args):
        self.modelXbrl.modelManager.cntlr.currentView = self

    def close(self):
        del self.viewFrame.view
        if self.modelXbrl:
            self.tabWin.forget(self.viewFrame)
            self.modelXbrl.views.remove(self)
            self.modelXbrl = None
            self.view = None

    def select(self):
        self.tabWin.select(self.viewFrame)

    def leave(self, *args):
        self.toolTipColId = None
        self.toolTipRowId = None

    def motion(self, *args):
        tvColId = self.treeView.identify_column(args[0].x)
        tvRowId = self.treeView.identify_row(args[0].y)
        if tvColId != self.toolTipColId or tvRowId != self.toolTipRowId:
            self.toolTipColId = tvColId
            self.toolTipRowId = tvRowId
            newValue = self.getToolTip(tvRowId, tvColId)
            if newValue is None and tvRowId and len(tvRowId) > 0:
                try:
                    col = int(tvColId[1:])
                    if col == 0:
                        newValue = self.treeView.item(tvRowId,"text")
                    else:
                        values = self.treeView.item(tvRowId,"values")
                        if col <= len(values):
                            newValue = values[col - 1]
                except ValueError:
                    pass
            self.setToolTip(newValue, tvColId)

    def getToolTip(self, rowId, colId):
        return None

    def setToolTip(self, text, colId="#0"):
        self.toolTip._hide()
        if isinstance(text,str) and len(text) > 0:
            width = self.treeView.column(colId,"width")
            if len(text) * 8 > width or '\n' in text:
                self.toolTipText.set(text)
                self.toolTip.configure(state="normal")
                self.toolTip._schedule()
            else:
                self.toolTipText.set("")
                self.toolTip.configure(state="disabled")
        else:
            self.toolTipText.set("")
            self.toolTip.configure(state="disabled")

    def contextMenu(self):
        try:
            return self.menu
        except AttributeError:
            try:
                self.menu = Menu( self.viewFrame, tearoff = 0 )
                self.treeView.bind( self.modelXbrl.modelManager.cntlr.contextMenuClick, self.popUpMenu, '+' )
                return self.menu
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating context menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None
                return None

    def popUpMenu(self, event):
        if self.menu:
            self.menuRow = self.treeView.identify_row(event.y)
            self.menuCol = self.treeView.identify_column(event.x)
            self.menu.post( event.x_root, event.y_root )

    def expand(self):
        self.setTreeItemOpen(self.menuRow,open=True)

    def expandAll(self):
        self.setTreeItemOpen("",open=True)

    def collapse(self):
        self.setTreeItemOpen(self.menuRow,open=False)

    def collapseAll(self):
        self.setTreeItemOpen("",open=False)

    def setTreeItemOpen(self, node, open=True):
        if node:
            self.treeView.item(node, open=open)
        for childNode in self.treeView.get_children(node):
            self.setTreeItemOpen(childNode, open)

    def menuAddExpandCollapse(self):
        if self.menu:
            self.menu.add_command(label=_("Expand"), underline=0, command=self.expand)
            self.menu.add_command(label=_("Collapse"), underline=0, command=self.collapse)
            self.menu.add_command(label=_("Expand all"), underline=0, command=self.expandAll)
            self.menu.add_command(label=_("Collapse all"), underline=0, command=self.collapseAll)

    def menuAddClipboard(self):
        if self.menu and self.modelXbrl.modelManager.cntlr.hasClipboard:
            try:
                clipboardMenu = Menu(self.viewFrame, tearoff=0)
                clipboardMenu.add_command(label=_("Cell"), underline=0, command=self.copyCellToClipboard)
                clipboardMenu.add_command(label=_("Row"), underline=0, command=self.copyRowToClipboard)
                clipboardMenu.add_command(label=_("Table"), underline=0, command=self.copyTableToClipboard)
                self.menu.add_cascade(label=_("Copy to clipboard"), menu=clipboardMenu, underline=0)
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating clipboard menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None

    def menuAddLangs(self):
        if self.menu:
            try:
                langsMenu = Menu(self.viewFrame, tearoff=0)
                self.menu.add_cascade(label=_("Language"), menu=langsMenu, underline=0)
                for lang in sorted(self.modelXbrl.langs):
                    langsMenu.add_command(label=lang, underline=0, command=lambda l=lang: self.setLang(l))
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating context languages menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None

    def menuAddLabelRoles(self, includeConceptName=False, menulabel=None):
        if self.menu:
            try:
                if menulabel is None: menulabel = _("Label role")
                rolesMenu = Menu(self.viewFrame, tearoff=0)
                self.menu.add_cascade(label=menulabel, menu=rolesMenu, underline=0)
                from arelle.ModelRelationshipSet import labelroles
                for x in labelroles(self.modelXbrl, includeConceptName):
                    rolesMenu.add_command(label=x[0][1:], underline=0, command=lambda a=x[1]: self.setLabelrole(a))
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating context label roles menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None

    def menuAddNameStyle(self, menulabel=None):
        if self.menu:
            try:
                if menulabel is None: menulabel = _("Name Style")
                nameStyleMenu = Menu(self.viewFrame, tearoff=0)
                self.menu.add_cascade(label=menulabel, menu=nameStyleMenu, underline=0)
                from arelle.ModelRelationshipSet import labelroles
                nameStyleMenu.add_command(label=_("Prefixed"), underline=0, command=lambda a=True: self.setNamestyle(a))
                nameStyleMenu.add_command(label=_("No prefix"), underline=0, command=lambda a=False: self.setNamestyle(a))
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating context name style menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None

    def menuAddUnitDisplay(self):
        if self.menu:
            try:
                rolesMenu = Menu(self.viewFrame, tearoff=0)
                self.menu.add_cascade(label=_("Units"), menu=rolesMenu, underline=0)
                rolesMenu.add_command(label=_("Unit ID"), underline=0, command=lambda: self.setUnitDisplay(unitDisplayID=True))
                rolesMenu.add_command(label=_("Measures"), underline=0, command=lambda: self.setUnitDisplay(unitDisplayID=False))
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating context unit menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None

    def menuAddViews(self, addClose=True, tabWin=None):
        if self.menu:
            try:
                if tabWin is None: tabWin = self.tabWin
                viewMenu = Menu(self.viewFrame, tearoff=0)
                self.menu.add_cascade(label=_("View"), menu=viewMenu, underline=0)
                newViewsMenu = Menu(self.viewFrame, tearoff=0)
                if addClose:
                    viewMenu.add_command(label=_("Close"), underline=0, command=self.close)
                viewMenu.add_cascade(label=_("Additional view"), menu=newViewsMenu, underline=0)
                newViewsMenu.add_command(label=_("Arcrole group..."), underline=0, command=lambda: self.newArcroleGroupView(tabWin))
                from arelle.ModelRelationshipSet import baseSetArcroles
                for x in baseSetArcroles(self.modelXbrl) + [( " Role Types","!CustomRoleTypes!"), (" Arcrole Types", "!CustomArcroleTypes!")]:
                    newViewsMenu.add_command(label=x[0][1:], underline=0, command=lambda a=x[1]: self.newView(a, tabWin))
            except Exception as ex: # tkinter menu problem maybe
                self.modelXbrl.info("arelle:internalException",
                                    _("Exception creating context add-views menu in %(title)s: %(error)s"),
                                    modelObject=self.modelXbrl.modelDocument, title=self.tabTitle, error=str(ex))
                self.menu = None

    def newView(self, arcrole, tabWin):
        if arcrole in ("!CustomRoleTypes!", "!CustomArcroleTypes!"):
            from arelle import ViewWinRoleTypes
            ViewWinRoleTypes.viewRoleTypes(self.modelXbrl, tabWin, arcrole=="!CustomArcroleTypes!", lang=self.lang)
        else:
            from arelle import ViewWinRelationshipSet
            ViewWinRelationshipSet.viewRelationshipSet(self.modelXbrl, tabWin, arcrole, lang=self.lang)

    def newArcroleGroupView(self, tabWin):
        from arelle.DialogArcroleGroup import getArcroleGroup
        from arelle import ViewWinRelationshipSet
        arcroleGroup = getArcroleGroup(self.modelXbrl.modelManager.cntlr, self.modelXbrl)
        if arcroleGroup:
            ViewWinRelationshipSet.viewRelationshipSet(self.modelXbrl, tabWin, arcroleGroup, lang=self.lang)

    def setLang(self, lang):
        self.lang = lang
        self.view()

    def setLabelrole(self, labelrole):
        self.labelrole = labelrole
        self.view()

    def setNamestyle(self, isPrefixed):
        self.nameIsPrefixed = isPrefixed
        self.view()

    def setUnitDisplay(self, unitDisplayID=False):
        self.unitDisplayID = unitDisplayID
        self.view()

    def setColumnsSortable(self, treeColIsInt=False, startUnsorted=False, initialSortCol="#0", initialSortDirForward=True):
        if hasattr(self, 'lastSortColumn') and self.lastSortColumn:
            self.treeView.heading(self.lastSortColumn, image=self.sortImages[2])
        self.lastSortColumn = None if startUnsorted else initialSortCol
        self.lastSortColumnForward = initialSortDirForward
        self.treeColIsInt = treeColIsInt
        if not hasattr(self, "sortImages"):
            self.sortImages = (PhotoImage(file=os.path.join(self.modelXbrl.modelManager.cntlr.imagesDir, "columnSortUp.gif")),
                               PhotoImage(file=os.path.join(self.modelXbrl.modelManager.cntlr.imagesDir, "columnSortDown.gif")),
                               PhotoImage())
        for col in ("#0",) + self.treeView["columns"]:
            self.treeView.heading(col, command=lambda c=col: self.sortColumn(c))
        if not startUnsorted:
            self.treeView.heading(initialSortCol, image=self.sortImages[not initialSortDirForward])

    def colSortVal(self, node, col):
        if col == "#0":
            treeColVal = self.treeView.item(node)["text"]
            if self.treeColIsInt:
                return int(treeColVal)
        else:
            treeColVal = self.treeView.set(node, col)
            if col == "sequence":
                try:
                    return int(treeColVal)
                except:
                    return 0
        return treeColVal

    def sortNestedRows(self, parentNode, col, reverse):
        l = [(self.colSortVal(node, col), node) for node in self.treeView.get_children(parentNode)]
        l.sort(reverse=reverse)
        # rearrange items in sorted positions
        for i, (cell, node) in enumerate(l):
            self.treeView.move(node, parentNode, i)
        # reset even/odd tags
        for i, node in enumerate(self.treeView.get_children(parentNode)):
            self.treeView.item(node, tags=('even' if i & 1 else 'odd',))
            self.sortNestedRows(node, col, reverse)

    def sortColumn(self, col):
        if col == self.lastSortColumn:
            reverse = self.lastSortColumnForward
            self.lastSortColumnForward = not reverse
        else:
            if self.lastSortColumn:
                self.treeView.heading(self.lastSortColumn, image=self.sortImages[2])
            reverse = False
            self.lastSortColumnForward = True
            self.lastSortColumn = col
        self.treeView.heading(col, image=self.sortImages[reverse])
        self.sortNestedRows('', col, reverse)
        self.viewSelectionChange()  # reselect selected rows

    def copyCellToClipboard(self, *ignore):
        self.modelXbrl.modelManager.cntlr.clipboardData(
            text=self.treeView.item(self.menuRow)['text'] if self.menuCol == '#0' else self.treeView.set(self.menuRow,self.menuCol))

    def copyRowToClipboard(self, *ignore):
        self.modelXbrl.modelManager.cntlr.clipboardData(
            text='\t'.join([self.treeView.item(self.menuRow)['text']] +
                           [self.treeView.set(self.menuRow,c) for c in self.treeView['columns']]))

    def copyTableToClipboard(self, *ignore):
        cols = self.treeView['columns']
        lines = ['\t'.join([self.treeView.heading('#0')['text']] +
                           [self.treeView.heading(h)['text'] for h in cols])]
        self.tabLines('', '', cols, lines)
        self.modelXbrl.modelManager.cntlr.clipboardData(text='\n'.join(lines))

    def tabLines(self, parentNode, indent, cols, lines):
        for node in self.treeView.get_children(parentNode):
            lines.append('\t'.join('"{}"'.format(c.replace('"','""')) if (isinstance(c, str) and "\n" in c) else c
                                   for c in ([indent + self.treeView.item(node)['text']] +
                                             [self.treeView.set(node,c) for c in cols])))
            self.tabLines(node, indent+'    ', cols, lines)



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import Toplevel, N, S, E, W, messagebox
try:
    from tkinter.ttk import Frame, Button
except ImportError:
    from ttk import Frame, Button
import regex as re
from arelle.ModelInstanceObject import NewFactItemOptions
from arelle.ModelValue import dateTime
from arelle import XmlUtil
from arelle.UiUtil import gridCell, gridCombobox, label
from arelle.CntlrWinTooltip import ToolTip

'''
caller checks accepted, if True, caller retrieves url
'''
def getNewFactItemOptions(mainWin, newInstanceOptions=None):
    if newInstanceOptions is None: newInstanceOptions = NewFactItemOptions()
    # use prior prevOptionValues for those keys not in existing newInstanceOptions
    for prevOptionKey, prevOptionValue in mainWin.config.get("newFactItemOptions",{}).items():
        if not getattr(newInstanceOptions, prevOptionKey, None):
            newInstanceOptions.__dict__[prevOptionKey] = prevOptionValue
    dialog = DialogNewFactItemOptions(mainWin, newInstanceOptions)
    if dialog.accepted:
        if dialog.options is not None: # pickle as strings, DateTime won't unpickle right
            mainWin.config["newFactItemOptions"] = dialog.options.__dict__.copy()
        mainWin.saveConfig()
        return True
    return False

monetaryUnits = ("AED", "AFN", "ALL", "AMD", "ANG", "AON", "ARP", "ATS", "AUD", "AWF", "AZM",
                 "AZN", "BAK", "BBD", "BDT", "BEF", "BGL", "BHD", "BIF", "BMD", "BND", "BOB",
                 "BRR", "BSD", "BTR", "BWP", "BYR", "BZD", "CAD", "CHF", "CLP", "CNY", "COP",
                 "CRC", "CSK", "CUP", "CVE", "CYP", "DEM", "DJF", "DKK", "DOP", "DZD", "ECS",
                 "EEK", "EGP", "ERN", "ESP", "ETB", "EUR", "FIM", "FJD", "FKP", "FRF", "GBP",
                 "GEL", "GHC", "GIP", "GMD", "GNF", "GRD", "GTQ", "GYD", "HKD", "HNL", "HRK",
                 "HTG", "HUF", "IDR", "IEP", "IEP", "ILS", "INR", "IQD", "IRR", "ISK", "ITL",
                 "JMD", "JOD", "JPY", "KES", "KGS", "KHR", "KMF", "KPW", "KRW", "KWD", "KYD",
                 "KZT", "LAK", "LBP", "LKR", "LRD", "LSL", "LTL", "LUF", "LVL", "LYD", "MAD",
                 "MDL", "MGF", "MKD", "MMK", "MNT", "MOP", "MRO", "MTL", "MUR", "MVR", "MWK",
                 "MXP", "MYR", "MZM", "NAD", "NGN", "NIO", "NLG", "NOK", "NPR", "NZD", "OMR",
                 "PAB", "PEN", "PGK", "PHP", "PKR", "PLN", "PTE", "PYG", "QAR", "ROL", "RSD",
                 "RUR", "RWF", "SAR", "SBD", "SBL", "SCR", "SDD", "SEK", "SGD", "SHP", "SIT",
                 "SKK", "SLL", "SOS", "SRG", "STD", "SVC", "SYP", "SZL", "THB", "TJR", "TMM",
                 "TND", "TOP", "TRL", "TTD", "TWD", "TZS", "UAH", "UGX", "USD", "UYU", "UZS",
                 "VEB", "VND", "VUV", "WST", "XAF", "XAG", "XAU", "XCD", "XDR", "XOF", "XPD",
                 "XPF", "XPT", "YER", "YUN", "ZAR", "ZMK", "ZRN", "ZWD",)

decimalsPattern = re.compile(r"^(INF|-?[0-9]+)$")

class DialogNewFactItemOptions(Toplevel):
    def __init__(self, mainWin, options):
        self.mainWin = mainWin
        parent = mainWin.parent
        super(DialogNewFactItemOptions, self).__init__(parent)
        self.parent = parent
        self.options = options
        parentGeometry = re.match("(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.accepted = False

        self.transient(self.parent)
        self.title(_("New Fact Item Options"))

        frame = Frame(self)

        label(frame, 1, 1, "Entity scheme:")
        self.cellEntityIdentScheme = gridCell(frame, 2, 1, getattr(options,"entityIdentScheme",""), width=50)
        ToolTip(self.cellEntityIdentScheme, text=_("Enter the scheme for the context entity identifier"), wraplength=240)
        label(frame, 1, 2, "Entity identifier:")
        self.cellEntityIdentValue = gridCell(frame, 2, 2, getattr(options,"entityIdentValue",""))
        ToolTip(self.cellEntityIdentValue, text=_("Enter the entity identifier value (e.g., stock ticker)"), wraplength=240)
        label(frame, 1, 3, "Start date:")
        self.cellStartDate = gridCell(frame, 2, 3, getattr(options,"startDate",""))
        ToolTip(self.cellStartDate, text=_("Enter the start date for the report period (e.g., 2010-01-01)"), wraplength=240)
        label(frame, 1, 4, "End date:")
        self.cellEndDate = gridCell(frame, 2, 4, getattr(options,"endDate",""))
        ToolTip(self.cellEndDate, text=_("Enter the end date for the report period (e.g., 2010-12-31)"), wraplength=240)
        label(frame, 1, 5, "Monetary unit:")
        self.cellMonetaryUnit = gridCombobox(frame, 2, 5, getattr(options,"monetaryUnit",""), values=monetaryUnits)
        ToolTip(self.cellMonetaryUnit, text=_("Select a monetary unit (e.g., EUR)"), wraplength=240)
        label(frame, 1, 6, "Monetary decimals:")
        self.cellMonetaryDecimals = gridCell(frame, 2, 6, getattr(options,"monetaryDecimals","2"))
        ToolTip(self.cellMonetaryDecimals, text=_("Enter decimals for monetary items"), wraplength=240)
        label(frame, 1, 7, "Non-monetary decimals:")
        self.cellNonMonetaryDecimals = gridCell(frame, 2, 7, getattr(options,"nonMonetaryDecimals","0"))
        ToolTip(self.cellNonMonetaryDecimals, text=_("Enter decimals for non-monetary items (e.g., stock shares)"), wraplength=240)

        cancelButton = Button(frame, text=_("Cancel"), width=8, command=self.close)
        ToolTip(cancelButton, text=_("Cancel operation, discarding changes and entries"))
        okButton = Button(frame, text=_("OK"), width=8, command=self.ok)
        ToolTip(okButton, text=_("Accept the options as entered above"))
        cancelButton.grid(row=8, column=1, columnspan=3, sticky=E, pady=3, padx=3)
        okButton.grid(row=8, column=1, columnspan=3, sticky=E, pady=3, padx=86)

        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(2, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))

        #self.bind("<Return>", self.ok)
        #self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def checkEntries(self):
        errors = []
        if not self.cellEntityIdentScheme.value:
            errors.append(_("Entity scheme invalid"))
        if not self.cellEntityIdentValue.value:
            errors.append(_("Entity identifier value invalid"))
        if not self.cellStartDate.value or dateTime(self.cellStartDate.value) is None:
            errors.append(_("Start date invalid"))
        if not self.cellEndDate.value or dateTime(self.cellEndDate.value) is None:
            errors.append(_("End date invalid"))
        if self.cellMonetaryUnit.value not in monetaryUnits:
            errors.append(_("Monetary unit invalid"))
        if not decimalsPattern.match(self.cellMonetaryDecimals.value):
            errors.append(_("Monetary decimals invalid"))
        if not decimalsPattern.match(self.cellNonMonetaryDecimals.value):
            errors.append(_("Non-monetary decimals invalid"))
        if errors:
            messagebox.showwarning(_("Dialog validation error(s)"),
                                "\n ".join(errors), parent=self)
            return False
        return True

    def setOptions(self):
        # set formula options
        self.options.entityIdentScheme = self.cellEntityIdentScheme.value
        self.options.entityIdentValue = self.cellEntityIdentValue.value
        # need datetime.datetime base class for pickling, not ModelValue class (unpicklable)
        self.options.startDate = self.cellStartDate.value
        self.options.endDate = self.cellEndDate.value
        self.options.monetaryUnit = self.cellMonetaryUnit.value
        self.options.monetaryDecimals = self.cellMonetaryDecimals.value
        self.options.nonMonetaryDecimals = self.cellNonMonetaryDecimals.value

    def ok(self, event=None):
        if not self.checkEntries():
            return
        self.setOptions()
        self.accepted = True
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()



'''
See COPYRIGHT.md for copyright information.
'''
from arelle import ModelObject, ModelDtsObject, XbrlConst, XmlUtil, ViewFile
from arelle.ModelDtsObject import ModelRelationship
from arelle.ViewUtil import viewReferences
import os

def viewRoleTypes(modelXbrl, outfile, header, isArcrole=False, lang=None):
    modelXbrl.modelManager.showStatus(_("viewing arcrole types") if isArcrole else _("viewing role types"))
    view = ViewRoleTypes(modelXbrl, outfile, header, isArcrole, lang)
    view.view()
    view.close()

class ViewRoleTypes(ViewFile.View):
    def __init__(self, modelXbrl, outfile, header, isArcrole, lang):
        super(ViewRoleTypes, self).__init__(modelXbrl, outfile, header, lang)
        self.isArcrole = isArcrole

    def view(self):
        # determine relationships indent depth for dimensions linkbases
        # set up treeView widget and tabbed pane
        if self.isArcrole:
            heading = ["Arcrole URI", "Definition", "Cycles Allowed", "Used On"]
            xmlRowElementName = "arcroleType"
            uriAttr = "arcroleURI"
        else:
            heading = ["Role URI", "Definition", "Used On"]
            xmlRowElementName = "roleType"
            uriAttr = "roleURI"
        self.addRow(heading, asHeader=True) # must do after determining tree depth

        roletypes = self.modelXbrl.arcroleTypes if self.isArcrole else self.modelXbrl.roleTypes
        for roleUri in sorted(roletypes.keys()):
            for modelRoleType in roletypes[roleUri]:
                attr = {"definedIn": modelRoleType.modelDocument.basename}
                cols = [roleUri, modelRoleType.genLabel(lang=self.lang, strip=True) or modelRoleType.definition or '']
                if self.isArcrole:
                    cols.append(modelRoleType.cyclesAllowed)
                cols.append(', '.join(str(usedOn)
                                      for usedOn in modelRoleType.usedOns))

                self.addRow(cols, treeIndent=0, xmlRowElementName=xmlRowElementName, xmlRowEltAttr=attr)
